# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Python implementation of the elevation meshing GUI."""

import os
from typing import Any, Dict, Iterable, Optional
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import (QCheckBox, QComboBox, QGroupBox, QLineEdit,
                             QPlainTextEdit, QPushButton, QRadioButton,
                             QSpinBox, QTextEdit)
from qgis.core import QgsMeshLayer, QgsProject  # type: ignore
from qgis.gui import QgisInterface  # type: ignore
from ...core import Lattice
from ...feedback import Feedback
from ...meshing import elevation_mesh
from ...types import Line3D, Point3D
from .. import utils
from .base import DialogBase


# NOTE: If the font "Courier New" is unavailable, Qt will pick a different
# monospace font based on the style hint.
_monospace_font = QFont('Courier New')
_monospace_font.setStyleHint(QFont.Monospace)


class ElevMeshDialog(DialogBase):
    """GUI implementation of the elevation meshing utility."""

    # NOTE: Qt widgets just have a ton of attributes - too bad!
    # pylint: disable=too-many-instance-attributes
    # Qt also uses camelCase for its attributes - too bad!
    # pylint: disable=invalid-name

    ui_file = 'ui_elevmesh.ui'
    help_file = 'help_elevmesh.md'

    def __init__(self, iface: QgisInterface) -> None:
        """Connect signal callbacks and populate widgets.

        Parameters
        ----------
        iface : QgisInterface
            The QGIS interface to create the dialog for
        """
        super().__init__(iface)  # Initialize DialogBase

        # NOTE: The following are attribute type hints used to inform supported
        # IDEs of the attribute types defined in the Qt UI file. This code is
        # executed during runtime, but does not do anything as far as the
        # Python interpreter is concerned.
        self.constraintsGroup: QGroupBox
        self.useLines: QCheckBox
        self.linesLayer: QComboBox
        self.usePoints: QCheckBox
        self.pointsLayer: QComboBox
        self.meshDomainGroup: QGroupBox
        self.keepConvexHull: QRadioButton
        self.shrinkToSegments: QRadioButton
        self.customBoundary: QRadioButton
        self.boundaryLayer: QComboBox
        self.settingsGroup: QGroupBox
        self.logLevel: QComboBox
        self.useSnapping: QCheckBox
        self.snappingTolerance: QSpinBox
        self.outputGroup: QGroupBox
        self.filePath: QLineEdit
        self.fileBrowseBtn: QPushButton
        self.keepTempFiles: QCheckBox
        self.addToMap: QCheckBox
        self.log: QPlainTextEdit
        self.saveLogBtn: QPushButton
        self.clearLogBtn: QPushButton
        self.help: QTextEdit
        self.cancelBtn: QPushButton
        self.runBtn: QPushButton
        self.closeBtn: QPushButton

        self.log.setFont(_monospace_font)
        self.load_help(self.help)
        self.status.setText('Ready')

        # Populate combo boxes
        self.linesLayer.clear()
        self.pointsLayer.clear()
        self.boundaryLayer.clear()
        self.add_polyline_layers(self.linesLayer, has_z=True)
        self.add_point_layers(self.pointsLayer, has_z=True)
        self.add_polygon_layers(self.boundaryLayer)

        # Hook up callbacks
        self.fileBrowseBtn.clicked.connect(
            self.browse_for_output('2dm', self.filePath, 'elevation-mesh'))
        self.saveLogBtn.clicked.connect(
            self.export_log('elevmesh', self.log))

        # Disabled features
        self.customBoundary.setEnabled(False)

    def accept(self) -> None:
        """Evaluate the user's chosen settings and run the tool.

        This function runs when the user confirms their settings, its
        name may not be changed as it is tied directly into the default
        QDialog slot.
        """
        kwargs: Dict[str, Any] = {}

        # Input geometry
        lines_data = self.linesLayer.currentData()
        if self.useLines.isChecked() and lines_data is not None:
            lines = list(utils.extract_lines(lines_data))
        else:
            lines = []
        points_data = self.pointsLayer.currentData()
        if self.usePoints.isChecked() and points_data is not None:
            points = list(utils.extract_points(points_data))
        else:
            points = []

        # Mesh domain
        if self.keepConvexHull.isChecked():
            kwargs['keep_convex_hull'] = True
        # Settings
        log_level = self.logLevel.currentIndex()
        kwargs['debug_level'] = log_level
        precision = 0.0
        if self.useSnapping.isChecked():
            precision = float(10 ** self.snappingTolerance.value())
        # Output
        filename = self.filePath.text()
        if self.keepTempFiles.isChecked():
            kwargs['move_tempfiles'] = os.path.dirname(filename)
        add_to_map = self.addToMap.isChecked()

        def after(_: Any, error: Optional[BaseException]) -> None:
            """Run after the worker has finished.

            Parameters
            ----------
            _ : Any
                The worker's return value; discarded and unneeded
            error : Optional[BaseException]
                The error that forced the end of execution, by default
                None
            """
            if error is None and add_to_map:
                layer_name = os.path.basename(filename).rsplit('.', 1)[0]
                layer = QgsMeshLayer(filename, layer_name, 'mdal')
                QgsProject.instance().addMapLayer(layer)
            self._set_lockout(False)
            if error is not None:
                self._push_status((0.0, 'Error', False))
                self.print_error(error)

        self._set_lockout(True)
        self.run_worker(precision, points, lines, filename,
                        log=self.log, after=after,
                        cancel_btn=self.cancelBtn, **kwargs)

    @staticmethod
    def worker(feedback: Feedback, precision: float, points: Iterable[Point3D],
               lines: Iterable[Line3D], filename: str, **kwargs: Any) -> None:
        """Perform the actual work of this utility.

        Any keyword arguments are passed on to the elevation_mesh
        function.

        Parameters
        ----------
        feedback : Feedback
            The feedback object use to communicate with the parent
            thread; this will be auto-filled internally
        precision : float
            The precision to use when generating the lattice
        points : Iterable[Tuple[float, float, float]]
            Input points to triangulate
        lines : Iterable[Tuple[float, float, float],
                         Tuple[float, float, float]]
            Input break lines to triangulate
        filename : str
            The output path of the generated mesh
        """

        def process_input_data(precision: float, points: Iterable[Point3D],
                               lines: Iterable[Line3D], feedback: Feedback
                               ) -> Lattice:
            """Consolidate the given input geometry into a lattice.

            Parameters
            ----------
            precision : float
                The precision to use when generating the lattice
            points : Iterable[Tuple[float, float, float]]
                Points to add to the lattice
            lines : Iterable[Tuple[float, float, float],
                            Tuple[float, float, float]]
                Lines to add to the lattice
            feedback : Feedback
                The feedback object used to communicate with the parent
                thread

            Returns
            -------
            Lattice
                A lattice containing the processed input data
            """
            # Instantiate lattice. This object will hold and process all of the
            # input geometries.
            lattice = Lattice(precision=precision)

            # Load input points into lattice
            with feedback.busy('Adding geometries to lattice...'):
                print('Reading points...')
                for point in points:
                    lattice.add_node(point, auto_conform=False)
                # Load input line segments into lattice
                print('Reading break lines...')
                for line in lines:
                    nodes = [lattice.add_node(p, auto_conform=False)
                             for p in line]
                    lattice.add_segment(nodes[0], nodes[1], auto_conform=False)

            # Conform input data. This removes duplicates, splits segments, etc.
            print('Conforming input geometries...')
            lattice.conform(feedback=feedback)

            return lattice

        print('Processing input geometries\n---------------------------')
        feedback.set_prefix('Step 1 of 2: ')
        lattice = process_input_data(precision, points, lines, feedback)

        feedback.set_prefix('Step 2 of 2: ')
        print('\n\nRunning Triangle\n----------------\n')
        with feedback.busy('Running Triangle'):
            mesh = elevation_mesh(*lattice.as_pslg(), **kwargs)

        mesh.save(filename, num_materials=0)
        feedback.clear_prefix()
        feedback.clear_scaling()
        feedback.update(1.0, 'Done')

    def _set_lockout(self, is_locked: bool) -> None:
        """Lock out the GUI while running the tool.

        Parameters
        ----------
        is_locked : bool
            Whether to lock or unlock the GUI
        """
        self.constraintsGroup.setEnabled(not is_locked)
        self.meshDomainGroup.setEnabled(not is_locked)
        self.settingsGroup.setEnabled(not is_locked)
        self.outputGroup.setEnabled(not is_locked)
        self.closeBtn.setEnabled(not is_locked)
        self.runBtn.setEnabled(not is_locked)
        self.cancelBtn.setEnabled(is_locked)



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Helper classes and methods used to provide LRO feedback."""

from types import TracebackType
from typing import Callable, Optional, Type

# Type aliases
FeedbackCallback = Callable[[Optional[float], Optional[str], bool], None]


class Feedback:
    """Generic interface for interacting with long-running functions.

    This class allows functions performing long-running operations
    (aka. LROs, generally anything over 1-2 seconds of blocking time)
    to communicate with their caller.

    The function can share its currents status using the
    Feedback.update() method. For processes of unknown length (such as
    network activity or external programs), you can use the
    Feedback.busy() context manager.

    If your workload supports cancellation, be sure to check the
    Feedback.is_cancelled flag as often as is sensible. It is set when
    the caller requests the termination of the operation.

    The callback function provided as part of the initialiser will be
    passed three positional arguments: a float between 0.0 and 1.0 that
    signifies the progress of the function, a status string, and a
    Boolean flag that shows whether the function flagged itself as busy
    (i.e. working but unable to give progress updates).
    """

    class FeedbackBusyContext:
        """Context manager for unresponsive subtask states.

        This only sets the Feedback.is_busy flag and is provided for
        syntactic sugar. It performs no error handling or suppression
        whatsoever.
        """

        def __init__(self, feedback: 'Feedback',
                     msg: Optional[str] = None) -> None:
            """Create the feedback context.

            Parameters
            ----------
            feedback : Feedback
                The feedback object to modify.
            msg : str, optional
                A status message to broadcast when entering the context
                manager, by default None
            """
            self.feedback = feedback
            self._msg = msg

        def __enter__(self) -> None:
            """Flag the feedback object as busy.

            This is called when the feedback.busy() context handler
            is entered. Does not return a value.
            """
            self.feedback._is_busy = True
            msg = None
            if self._msg is not None:
                msg = self.feedback._prefix + self._msg
            self.feedback._callback(None, msg, True)

        def __exit__(self, exc_type: Type[BaseException],  # type: ignore
                     exc_val: BaseException,
                     exc_traceback: TracebackType) -> bool:
            """Flag the feedback object as no longer busy.

            This performs no error handling whatsoever.

            Parameters
            ----------
            exc_type : Type[BaseException]
                The exception class that was raised
            exc_val : BaseException
                The exception instance raised
            exc_traceback : TracebackType
                Exception traceback

            Returns
            -------
            bool
                Whether the exception should be suppressed
            """
            self.feedback._is_busy = False
            return False

    def __init__(self, callback: FeedbackCallback,
                 prefix: Optional[str] = None) -> None:
        """Set up a new feedback wrapper using the given callback.

        The callback function will be called whenever the status is
        updated, make sure it is very light-weight or rate limited if
        necessary.

        See the docstring of the Feedback class for a list of arguments
        that will be passed.

        Parameters
        ----------
        callback : Callable[Optional[float], Optional[str], bool]
            The callback function to run when the status is updated
        prefix : str, optional
            An optional prefix to use for status messages; reset using
            the Feedback.set_prefix() method, by default None
        """
        self._callback = callback
        self._is_busy = False
        self._is_cancelled = False
        self._prefix = '' if prefix is None else prefix
        self._scaling_span = 1.0
        self._scaling_offset = 0.0

    @property
    def is_busy(self) -> bool:
        """Return whether the task is flagged as busy."""
        return self._is_busy

    @property
    def is_cancelled(self) -> bool:
        """Return whether the task has been cancelled."""
        return self._is_cancelled

    def busy(self, msg: Optional[str] = None) -> 'FeedbackBusyContext':
        """Return a context manager for the FEedback.is_busy flag.

        Use this to flag a part of your LRO as busy, meaning that it is
        working but unable to give progress updates, as may occur while
        waiting for an external executable to exit, or while awaiting
        network traffic.::

            with feedback.busy('Awaiting execution...'):
                # Within this block, the feedback status will
                # always report "busy" and any updates to the
                # progress will be silently ignored.
                subtask.part1()
                # Updating the status message is still permissable
                feedback.update(msg='Almost done...')
                subtask.part2(b)
            feedback.update(1.0, 'Done')

        Parameters
        ----------
        msg : str, optional
            A status message to when the context manager is entered, by
            default None

        Returns
        -------
        FeedbackBusyContext
            A context manager during which the feedback object will be
            flagged as busy

        """
        context = self.FeedbackBusyContext(self, msg)
        return context

    def cancel(self) -> None:
        """Flag the current task as cancelled.

        This informs the function that it should exit gracefully as
        quickly as possible.
        """
        self._is_cancelled = True

    def clear_prefix(self) -> None:
        """Clear the status message prefix."""
        self._prefix = ''

    def clear_scaling(self) -> None:
        """Reset the scaling behaviour."""
        self._scaling_span = 1.0
        self._scaling_offset = 0.0

    @staticmethod
    def scale(value: float, total: float, offset: float = 0.0) -> float:
        """Rescale a given progress value to another range.

        This is mostly used to calculate the total progress of multiple
        sub-tasks, e.g. being 80% finished with a task that is itself
        only half the total workload would mean 40% total progress:

            >>> Feedback.scale(0.8, 0.5)
            0.4

        Parameters
        ----------
        value : float
            The progress of the sub task
        total : float
            How big the sub task is relative to the total job
        offset : float, optional
            An offset to prepend; useful to add the weight of other
            sub tasks that already completed, by default 0.0

        Returns
        -------
        float
            The scaled progress value
        """
        if value < 0.0:
            value = 0.0
        elif value > 1.0:
            value = 1.0
        if total < 0.0:
            total = 0.0
        elif total > 1.0 - offset:
            total = 1.0 - offset
        new_val = offset + value*total
        if new_val < 0.0:
            return 0.0
        if new_val > 1.0:
            return 1.0
        return new_val

    def set_busy(self, is_busy: bool = True) -> None:
        """Set the Feedback.is_busy flag.

        This function is provided for compatibility, it is recommended
        to use the Feedback.busy() context manager instead. See its
        docstring for details.

        Parameters
        ----------
        is_busy : bool, optional
            Whether to flag the feedback object as busy, by default
            True
        """
        self._is_busy = is_busy

    def set_prefix(self, prefix: str) -> None:
        """Set the status message prefix.

        Use Feedback.clear_prefix() to remove the prefix entirely.

        Parameters
        ----------
        prefix : str
            A new prefix to prepend to every status message sent
        """
        self._prefix = prefix

    def set_scaling(self, span: float, offset: float = 0.0) -> None:
        """Set the scaling factor.

        The scaling factor automatically applies the scale() method
        to any status updates.

        Parameters
        ----------
        span : float
            The scaling factor for any progress value passed
        offset : float, optional
            An offset to account for past operations, by default 0.0

        Raises
        ------
        ValueError
            Raised if span is less than 0 or greater than 1
        ValueError
            Raised if offset is less than 0 or greater than 1-span
        """
        if not 0.0 <= span <= 1.0:
            raise ValueError('span must lie within 0.0 and than 1.0')
        if not 0.0 <= offset <= 1.0-span:
            raise ValueError('offset must lie within 0.0 and (1.0 - span)')
        self._scaling_span = span
        self._scaling_offset = offset

    def update(self, value: Optional[float] = None,
               msg: Optional[str] = None) -> None:
        """Push the given status and progress.

        This is the primary endpoint for functions reporting status
        updates to their callers.

        None values mean that the value should not be updated

        Parameters
        ----------
        value : float, optional
            A value from 0 to 1 that reflects the aproximate completion
            status of the operation, by default None
        msg : str, optional
            A status message to push, by default None

        Raises
        ------
        InterruptedError
            Raised if the cancel flag was set
        ValueError
            Raised if value is less than 0 or greater than 1
        """
        if self._is_cancelled:
            self._callback(0.0, 'Operation cancelled', False)
            raise InterruptedError('The operation was cancelled by the user')
        if msg is not None:
            msg = self._prefix + msg
        if value is not None and not self._is_busy:
            if not 0.0 <= value <= 1.0:
                raise ValueError('value must lie within 0.0 and 1.0')
            value = self.scale(value, self._scaling_span, self._scaling_offset)
        self._callback(value, msg, self._is_busy)



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""QGIS plugin component of BASEmesh."""

from typing import Optional
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QAction, QMenu, QToolBar
from qgis.core import QgsApplication  # type: ignore
from qgis.gui import QgisInterface  # type: ignore
from . import resources
from .gui import AboutDialog, ElevMeshDialog, InterpolDialog, QualMeshDialog
from .processing import BasemeshProcessingProvider


class BaseMeshPlugin:
    """The BASEmesh plugin to be loaded into the QGIS application."""

    display_name = 'BASEmesh 2'

    def __init__(self, iface: QgisInterface) -> None:
        """Prepare the plugin for initialisation.

        Note that this is called in preparation for the plugin being
        loaded. The class may be initialised, but the QGIS plugin side
        still requires the initGui() method to be called.

        Parameters
        ----------
        iface : QgisInterface
            The interface to generate the plugin's menu for
        """
        self.iface = iface
        self.menu: Optional[QMenu] = None
        self.toolbar: Optional[QToolBar] = None
        self.provider = BasemeshProcessingProvider()
        # Create actions
        self.about = QAction(
            QIcon(':/plugins/basemesh/plugin/icons/basemesh_icon.svg'),
            'About BASEmesh')
        self.about.setStatusTip(
            'Additional information about BASEmesh, BASEMENT, and Triangle')
        self.about.triggered.connect(self._open_about)
        self.elevmesh = QAction(
            QIcon(':/plugins/basemesh/plugin/icons/elev_meshing.svg'),
            'Elevation meshing')
        self.elevmesh.setStatusTip('Meshing of elevation data')
        self.elevmesh.triggered.connect(self._open_elevmesh)
        self.qualmesh = QAction(
            QIcon(':/plugins/basemesh/plugin/icons/qual_meshing.svg'),
            'Quality meshing')
        self.qualmesh.setStatusTip('Meshing with quality constraints')
        self.qualmesh.triggered.connect(self._open_qualmesh)
        self.interpol = QAction(
            QIcon(':/plugins/basemesh/plugin/icons/interpolation.svg'),
            'Interpolation')
        self.interpol.setStatusTip('Mesh interpolation')
        self.interpol.triggered.connect(self._open_interpol)

    def initGui(self) -> None:
        """GUI initialization procedure.

        This method is called by QGIS to integrate the plugin with the
        QGIS user interface. This includes menu bar items and buttons.
        """
        assert self.toolbar is not None
        toolbar: QToolBar = self.toolbar
        # Create toolbar
        toolbar = self.iface.addToolBar(self.display_name)
        toolbar.setObjectName('basemeshToolBar')
        toolbar.addAction(self.elevmesh)
        toolbar.addAction(self.qualmesh)
        toolbar.addAction(self.interpol)

        # Add plugin menu entry
        assert self.menu is not None
        menu: QMenu = self.menu
        menu = self.iface.pluginMenu().addMenu(
            QIcon(':/plugins/basemesh/plugin/icons/basemesh_icon.svg'),
            self.display_name)
        menu.addAction(self.elevmesh)
        menu.addAction(self.qualmesh)
        menu.addAction(self.interpol)
        menu.addAction(self.about)

        # Hook up processing algorithm provider
        QgsApplication.processingRegistry().addProvider(self.provider)

    def _open_about(self) -> None:
        AboutDialog(self.iface).exec_()

    def _open_elevmesh(self) -> None:
        ElevMeshDialog(self.iface).exec_()

    def _open_interpol(self) -> None:
        InterpolDialog(self.iface).exec_()

    def _open_qualmesh(self) -> None:
        QualMeshDialog(self.iface).exec_()

    def unload(self) -> None:
        """GUI breakdown procedure.

        This method is called when the plugin is unloaded. After this
        function completes, the interface is expected to be as if the
        plugin were never initialized in the first place.
        """
        # Destroy plugin menu entry
        if self.menu is not None:
            self.menu.removeAction(self.elevmesh)
            self.menu.removeAction(self.interpol)
            self.menu.removeAction(self.qualmesh)
            self.menu.removeAction(self.about)
            self.iface.pluginMenu().removeAction(self.menu.menuAction())
        # Remove tool bar
        del self.toolbar

        # Remove processing algorithm provider
        QgsApplication.processingRegistry().removeProvider(self.provider)


def classFactory(iface: QgisInterface) -> 'BaseMeshPlugin':
    """Instantiate and return the BASEmesh plugin object.

    This function is called by QGIS  when loading the plugin and serves
    as the primary endpoint to the QGIS application.

    Parameters
    ----------
    iface : QgisInterface
        The QGIS interface to generate the plugin menu

    Returns
    -------
    BaseMeshPlugin
        The instantiated BASEmesh plugin to be used by QGIS
    """
    return BaseMeshPlugin(iface)



"""Installation script for BASEmesh.

This is used to set up the module after installation via PIP, as well
as to build local versions for testing and development.
"""

from setuptools import Extension, find_packages, setup

from basemesh import __version__

# Load the project README
with open('README.md') as readme:
    long_description = readme.read()

# Define the algorithm acceleration module
algorithms_c = Extension(name='basemesh.algorithms.algorithms_c',
                         sources=['src/algorithms_c.c'],
                         optional=True)

# Non-Python files to include
package_data = {
    'basemesh': ['metadata.txt', 'packages/**'],
    'basemesh.plugin': ['resources.qrc', 'icons/**'],
    'basemesh.plugin.gui': ['templates/**'],
    'basemesh.triangle': ['bin/**']
}

setup(name='BASEmesh',
      version=__version__,
      description='Pre-processing and mesh generation toolkit for BASEMENT.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='ETH Zürich',
      maintainer_email='seidelmann@vaw.ethz.ch',
      url='https://git.ee.ethz.ch/basemesh/basemesh-v2/',
      license='GNU General Public License v3 (GPLv3)',
      packages=find_packages(),
      package_data=package_data,
      ext_modules=[
          algorithms_c
      ],
      classifiers=[
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Programming Language :: Python :: 3.7',
          'Typing :: Typed'
      ],
      install_requires=[
          'numpy'
      ],
      scripts=['scripts/basechange.py'])



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Mesh interpolation tool."""

from typing import Callable, Dict, Optional, Tuple
from .abc import ElevationSource
from .core import Mesh
from .feedback import Feedback
from .types import Triangle2D, Point2D

# Type aliases
_TriangleSampler = Callable[[Triangle2D], Point2D]


def calculate_element_elevation(mesh: Mesh, *args: ElevationSource,
                                default: Optional[float] = None,
                                sampler: Optional[_TriangleSampler] = None,
                                feedback: Optional[Feedback] = None
                                ) -> Dict[int, float]:
    """Calculate and add the element elevation attributes to the mesh.

    This calculates mesh element elevations using a given element
    sampler. The mesh is only used to retrieve the 2D element
    positions, the node elevation is ignored when determining the
    element elevation.

    Parameters
    ----------
    mesh : Mesh
        The mesh to interpolate; its node elevation will be ignored
    *args : ElevationSource
        The elevation sources to use, in descending order of priority
    feedback : Feedback, optional
        A status feedback callable to use for the operation, by default
        None
    default : float, optional
        The default elevation if no value could be retrieved, by
        default None
    sampler : Callable[[Tuple[Tuple[float, float], Tuple[float, float],
                              Tuple[float, float]],
                        Tuple[float, float]]], optional
        The function to use to determine the sampling location for a
        given triangle, by default None

    Returns
    -------
    Dict[int, float]
        A mapping from element ID to elevation

    Raises
    ------
    ValueError
        Raised if no elevation sources were provided
    ValueError
        Raised if no elevation source could produce a value and no
        default value was specified
    """
    if len(args) < 1:
        raise ValueError('At least one elevation source must be provided')
    sources: Tuple[ElevationSource, ...] = tuple(args)

    def triangle_com(triangle: Triangle2D) -> Point2D:
        """Calculate a 2D triangle's centre of mass."""
        avg_x = (triangle[0][0] + triangle[1][0] + triangle[2][0]) / 3.0
        avg_y = (triangle[0][1] + triangle[1][1] + triangle[2][1]) / 3.0
        return avg_x, avg_y

    if sampler is None:
        sampler = triangle_com

    if feedback is not None:
        total = len(mesh.elements)
        interval = int(total / 100) if total > 100 else 1

    # Iterate over all elements in the input mesh
    result_dict: Dict[int, float] = {}
    for index, element in enumerate(mesh.elements):
        sample_point: Point2D = sampler(element.as_triangle_2d)
        height: Optional[float] = None

        # NOTE: This step is identical to the one in interpolate_mesh. It would
        # be nice to have a separate local elevation_at(<sources>) interface to
        # avoid this repetition.

        # Loop over all elevation sources
        for source in sources:
            try:
                # Try to get an elevation from the current source. The
                # ElevationSource ABC requires that a ValueError be raised if
                # the source is unable to produce a value.
                height = source.elevation_at(sample_point)
            except ValueError:
                # If no value could be returned, move on to the next elevation
                # source
                pass
            else:
                # The elevation source provided a height value, quit the loop
                break

        if height is None:
            # If no default height was provided, raise an error
            if default is None:
                raise ValueError(f'Unable to interpolate point {sample_point}')
            # Use the default height if provided
            height = default

        # NOTE: It would be possible to introduce an elevation attribute to the
        # Element object, but this would be confusing considering that node and
        # element elevations are not directly linked or inferrable.
        # To resolve this ambiguity, we would have to add a custom mesh
        # sub class that has only elevation attributes and no node elevation.
        #
        # However, due to the sampler system, this would tie the mesh object to
        # its defining elevation sources, so those would have to stick around
        # as long as the mesh does so we can re-calculate the elevations if the
        # nodes are moved.
        #
        # That too could be mitigated by making this mesh type immutable, but
        # at that point we'd lose mesh editing capability and add a lot of
        # complexity to the object model for no justifiable reason. Therefore,
        # we just calculate the element elevations into a dict and sprinkle
        # them into the 2DM during export.
        #
        # -- LS

        # Record the elevation for this point
        result_dict[element.id_] = height

        if feedback is not None and index % interval == 0:
            feedback.update((index+1)/total)

    return result_dict


def interpolate_mesh(mesh: Mesh, *args: ElevationSource,
                     default: Optional[float] = None,
                     feedback: Optional[Feedback] = None) -> None:
    """Interpolate a mesh using the provided elevation sources.

    This operation modifies the input mesh.

    Parameters
    ----------
    mesh : Mesh
        The mesh to interpolate the nodes of
    *args : ElevationSource
        Any number of elevation sources to use in descending order of
        priority, at least one is required
    default : float, optional
        A fallback value used when no elevation source could produce a
        value for a given point, by default None
    feedback : Feedback, optional
        A feedback callable to communicate with the caller, by default
        None

    Raises
    ------
    ValueError
        Raised if no elevation source has been specified.
    ValueError
        Raised if every elevation source has failed and no default
        value was provided
    """
    if len(args) < 1:
        raise ValueError('At least one elevation source must be provided')
    sources: Tuple[ElevationSource, ...] = tuple(args)

    if feedback is not None:
        total = len(mesh.nodes)
        interval = int(total / 100) if total > 100 else 1

    # Iterate over all nodes in the input mesh
    for index, node in enumerate(mesh.nodes):
        point_xy = node.as_tuple_2d()
        height: Optional[float] = None

        # Loop over all elevation sources
        for source in sources:
            try:
                # Try to get an elevation from the current source. The
                # ElevationSource ABC requires that a ValueError be raised if
                # the source is unable to produce a value.
                height = source.elevation_at(point_xy)
            except ValueError:
                # If no value could be returned, move on to the next elevation
                # source
                pass
            else:
                # The elevation source provided a height value, quit the loop
                break

        if height is None:
            # If no default height was provided, raise an error
            if default is None:
                raise ValueError(f'Unable to interpolate point {point_xy}')
            # Use the default height if provided
            height = default

        # Move the node to the new elevation
        node.move(z=height)

        if feedback is not None and index % interval == 0:
            feedback.update((index+1)/total)



"""Pre-processing and mesh generation toolkit for BASEMENT.

This package contains utilities for generating and processing 1D and 2D
mesh geometries for use with the numerical simulation software such as
BASEMENT (https://basement.ethz.ch/).

It features standalone mesh generation utilities leveraging Jonathan
Shewchunk's Triangle (https://www.cs.cmu.edu/~quake/triangle.html), as
well as mesh editing and interpolation utilities.

BASEmesh can also be installed as a plugin for QGIS
(https://qgis.org/en/site/), allowing usage of its functionality via
the plugin interface.

For additional information, refer to the BASEMENT website linked above
or visit the project repository at
https://git.ee.ethz.ch/basemesh/basemesh-v2/.


Copyright (C) 2020  ETH Zürich

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from . import abc
from . import meshtool
from . import triangle
from .core import Node, Mesh, MeshElement, MeshNode, Segment
from .interpolation import calculate_element_elevation, interpolate_mesh
from .meshing import elevation_mesh, quality_mesh
from .stringdefs import (resolve_string_defs, write_string_defs_mesh,
                         write_string_defs_sidecar)

__version__ = '2.0.4'

try:
    from qgis.gui import QgisInterface  # type: ignore
except ImportError:
    QGIS_ENVIRONMENT = False
else:
    QGIS_ENVIRONMENT = True
    from .plugin import classFactory



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Core components of the BASEmesh data model.

This includes the implementation of data classes for geometry and mesh
objects, as well as supporting code and utilities specific to them.
"""

from .geometry import Lattice, Node, Segment
from .mesh import Mesh, MeshElement, MeshNode

__all__ = [
    'Lattice',
    'Mesh',
    'MeshElement',
    'MeshNode',
    'Node',
    'Segment'
]



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Graphical user interface components for BASEmesh."""

from .about import AboutDialog
from .elevmesh import ElevMeshDialog
from .interpol import InterpolDialog
from .qualmesh import QualMeshDialog

__all__ = [
    'AboutDialog',
    'ElevMeshDialog',
    'InterpolDialog',
    'QualMeshDialog'
]



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Helper method for supporting rasterized elevation sources."""

import math
from qgis.core import QgsPointXY, QgsRasterLayer  # type: ignore
from ..abc import ElevationSource
from ..types import Point2D


class RasterDataWrapper(ElevationSource):
    """Wrapper class for QgsRasterLayer.

    This class allows usage of QgsRasterLayer as an interpolation
    source.
    """

    def __init__(self, layer: QgsRasterLayer, band_id: int = 0) -> None:
        """Instantiate the raster data wrapper.

        Note that this function uses a 0-based band index. The actual
        interpolation function uses a 1-based index, but this is
        converted internally.
        Be sure to pass the regular band index into this initialiser,
        i.e. the same you would use to retrieve the band name by index.

        Parameters
        ----------
        layer : QgsRasterLayer
            The raster layer to wrap
        band_id : int, optional
            The 0-based band index, by default 0
        """
        self.layer = layer
        self.band = band_id
        self.provider = layer.dataProvider()

    def elevation_at(self, point: Point2D) -> float:
        """Return the elevation at the given point.

        Parameters
        ----------
        point : Tuple[float, float]
            The point to interpolate

        Returns
        -------
        float
            The elevation of the data source at the given point

        Raises
        ------
        ValueError
            Raised if the data source could not provide a value for the
            given sample point
        """
        # Query the data provider for the elevation at the given point
        point_xy = QgsPointXY(*point)

        # NOTE: Important:
        # Although it is not currently marked as such in the documentation, the
        # QgsRasterDataProvider.sample() method is available to the Python
        # API.
        # Its arguments seem to be mostly the same, but the ok parameter has
        # been replaced with a second return value instead.
        #
        # Also, for reasons unknown, it uses band indices starting at 1.
        value, is_valid = self.provider.sample(point_xy, self.band+1)

        # If no match could be found, raise a ValueError
        if not is_valid or math.isnan(value):
            raise ValueError(f'Unable to interpolate point {point}')

        return float(value)



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""String definition parsing and writing."""

import os
from typing import Dict, List, Mapping, Optional
from .algorithms import dist_2d, point_on_line
from .core.mesh import Mesh, MeshNode
from .feedback import Feedback
from .types import LineString2D


def _group_nodestrings(string_defs: Mapping[str, LineString2D],
                       mesh: Mesh, precision: float = 0.0, *,
                       feedback: Optional[Feedback] = None) -> Dict[str, List[MeshNode]]:
    """Return a mapping from string definition names to nodes.

    This processes all nodes in the input mesh, checking for
    coincidence with all node strings. It generates a dictionary
    mapping the string definition names to a list of nodes that are
    within range of the string definition line string.

    Parameters
    ----------
    string_defs : Mapping[str, Tuple[Tuple[str, str], ...]]
        A mapping of string definitions to line strings
    mesh : Mesh
        The mesh to process
    precision : float, optional
        The maximum distance for a node to still be considered part of
        a line string, by default 0.0
    feedback : Feedback, optional
        A feedback object used to communicate with the caller, by
        default None

    Returns
    -------
    Dict[str, List[MeshNode]]
        A mapping from string definitions to mesh nodes
    """
    if feedback is not None:
        total = len(mesh.nodes)
        interval = int(total / 100) if total > 100 else 1

    nodes_dict: Dict[str, List[MeshNode]] = {n: [] for n in string_defs}
    for index, node in enumerate(mesh.nodes):
        node_pos = node.as_tuple_2d()
        for name, line_string in string_defs.items():

            # Check each subsegment
            for sub_index, point in enumerate(line_string[:-1]):
                line = point, line_string[sub_index+1]
                if point_on_line(line, node_pos, precision=precision):
                    nodes_dict[name].append(node)
                    break

        if feedback is not None and index % interval == 0:
            feedback.update(feedback.scale(index/total, 0.7))

    return nodes_dict


def resolve_string_defs(string_defs: Dict[str, LineString2D],
                        mesh: Mesh, precision: float = 0.0, *,
                        feedback: Optional[Feedback] = None) -> Dict[str, List[int]]:
    """Resolve the string definitions for the given mesh.

    This returns the list of mesh nodes lying on the given string, in
    order of distance to the first point in the line string.

    Parameters
    ----------
    string_defs : Dict[str, Tuple[Tuple[flost, float], ...]]
        The string definitions to resolve
    mesh : Mesh
        The mesh to process
    precision : float, optional
        The maximum distance for a node to still be considered part of
        a node string, by default 0.0
    feedback : Feedback, optional
        A feedback object used to communicate with the caller, by
        default None

    Returns
    -------
    Dict[str, List[int]]
        A mapping of string definitions to mesh node IDs
    """
    # Group nodes by their matching string definition line string
    nodes_dict: Dict[str, List[MeshNode]] = _group_nodestrings(
        string_defs, mesh, precision, feedback=feedback)

    # Sort string definition nodes based on distance to their start
    _sort_nodes(string_defs, nodes_dict, feedback=feedback)

    # Return string definition dict
    return {k: [n.id_ for n in v] for k, v in nodes_dict.items()}


def _sort_nodes(string_defs: Dict[str, LineString2D],
                nodes_dict: Dict[str, List[MeshNode]], *,
                feedback: Optional[Feedback] = None) -> None:
    """Sort the string definition nodes.

    When the nodes are assigned to their respective line strings(s),
    they are in an unsecified order as determined by the mesh node
    iterator.

    This function sorts them by distance from the first node in the
    string definition line string.

    This mutates the provided nodes_dict mapping.

    Parameters
    ----------
    string_defs : Dict[str, Tuple[Tuple[float, float], ...]]
        The original, ungrouped string definitions
    nodes_dict : Dict[str, List[MeshNode]]
        The grouped string definitions with associated mesh nodes
    feedback : Feedback, optional
        A feedback object used to communicate with the caller, by
        default None
    """
    if feedback is not None:
        total = len(nodes_dict)
        interval = int(total / 100) if total > 100 else 1

    for index, element in enumerate(nodes_dict.items()):
        name, nodes = element

        def rel_dist(node: MeshNode, sd_name: str = name) -> float:
            return dist_2d(string_defs[sd_name][0], node.as_tuple_2d())

        nodes_dict[name] = sorted(nodes, key=rel_dist)

        if feedback is not None and index % interval == 0:
            feedback.update(feedback.scale(index/total, 0.3, 0.7))


def write_string_defs_mesh(string_defs: Dict[str, List[int]],
                           file_path: str) -> None:
    """Append the string definitions into a 2DM mesh file.

    This format is required for BASEMENT 3.

    Parameters
    ----------
    string_defs : Dict[str, List[int]]
        A mapping from string definitions to mesh node IDs
    file_path : str
        The path of the mesh file to append

    Raises
    ------
    FileNotFoundError
        Raised if the mesh file could not be found
    """
    if not os.path.exists(file_path):
        raise FileNotFoundError('Unable to append string defs, path '
                                f'{file_path} does not exist')
    with open(file_path, 'a') as output:
        # HACK: Disabled readability newline for compatibility
        # output.write('\n')
        for name, node_ids in string_defs.items():
            # NOTE: According to the 2DM documentation, no more than 10 words
            # may follow the card identifier ("NS", in this case).
            # Longer node strings therefore have to be split up into multiple
            # cards. The actual end of a given node string is determined by a
            # negative node ID.
            # The insertion of the string definition name after the node string
            # end marker is out of spec.
            output.write('\nNS ')
            for index, node_id in enumerate(node_ids):
                # HACK: Disabled the max-10-tokens rule for BASEMENT
                # compatibility.
                # if (index+1) % 10 == 0:
                #     output.write('\nNS ')
                if index+1 == len(node_ids):
                    # HACK: Removed blank line between node strings
                    output.write(f'-{node_id} {name}')
                    break
                output.write(f'{node_id} ')


def write_string_defs_sidecar(string_defs: Dict[str, List[int]],
                              file_path: str) -> None:
    """Write the string definitions into a separate text file.

    This format is required for BASEMENT 2.8.

    Parameters
    ----------
    string_defs : Dict[str, List[int]]
        A mapping from string definitions to mesh node IDs
    file_path : str
        The path of the sidecar file to write
    """
    # Write string defs
    with open(file_path, 'w') as sd_file:
        for name, node_ids in string_defs.items():
            sd_file.write(f'STRINGDEF {{\n\tname = {name}\n\tnode_ids = (')
            sd_file.write(' '.join(str(i) for i in node_ids))
            sd_file.write(')\n\tupstream_direction = right\n}\n')



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""I/O module for 1D BASEchain geometry (BMG) files."""

import ast
import dataclasses
import datetime
import warnings
from types import TracebackType
from typing import (Any, Dict, IO, Iterable, Iterator, List, Optional,
                    Sequence, Tuple, Type, TYPE_CHECKING)

from ..errors import TokenError

if TYPE_CHECKING:
    from .geometry import ChannelGeometry


@dataclasses.dataclass()
class BaseChainSoilDef:
    """Data class for BMG "SOIL_DEF" blocks.

    Soil definition indices must be unique, and their ranges/indexes
    may not overlap.

    Attributes
    ----------
    index: int
        The unique ID of the soil definition
    range : Tuple[float, float], optional
        The range this soil definition applies to, by default None
        .. seealso:: :attr:`slice_indexes`
    slice_indexes: Tuple[int, int], optional
        Like :attr:`range`, but uses edge indexes instead of
        coordinates, by default None

    """

    index: int
    range: Optional[Tuple[float, float]] = None
    slice_indexes: Optional[Tuple[int, int]] = None


@dataclasses.dataclass()
class BaseChainCrossSection:  # pylint: disable=too-many-instance-attributes
    """Data class for BMG "CROSS_SECTION" blocks.

    Every element in the 1D computational grid is based around a cross
    section object, which describes the hydraulic and topological
    information for the element.

    Attributes
    ----------
    name : str
        The unique name of the cross section.
    distance_coord : float
        Distance coordinate of the cross section. This is measured from
        up- to downstream and positions the cross section. Please note
        that this value is given in kilometres.
    node_coords : List[Tuple[float, float]]
        The list of nodes defining the cross section geometry.

        This uses a relative coordinate system whose first element
        represents the distance from the leftmost node relative to the
        river flow direction. The second tuple element specifies the
        absolute elevation of the given node.
    sternberg : float, optional
        Sternberg-factor for the reduction of volume by abrasion
        ``V = V0e**(s*Dx)``, by default 0.0
    reference_height : float, optional
        A height offset to apply to any node-specific elevation values,
        by default 0.0
    left_point_global_coords : Tuple[float, float, float], optional
        Global coordinates of the leftmost cross section node. Allows
        converting the 1D channel back into a 2D geometry. In addition,
        it is used by the cross section interpolator to account for
        curvatures, etc., by default (0.0, 0.0, 0.0)
        .. seealso:: :attr:`orientation_angle`
    orientation_angle : float, optional
        In-plane angle of the cross section line with the leftmost node
        as the axis of rotation. The angle is measured counter
        clockwise with the vector ``(1,0)`` representing an angle of 0
        degrees, by default 0.0
        .. seealso:: :attr:`left_point_global_coords`
    interpolation_fixpoints : List[int], optional
        The interpolation fixpoints are used to specify break lines
        in-between existing cross sections. This can be used to insert
        new cross sections, as well as to improve the quality of a 2D
        mesh export.

        The first and last node in a cross section are always
        connected. Nodes are numbered starting at 1, by default None
    water_flow_range : Tuple[float, float], optional
        Used to limit the part of the part of the cross section that
        will be used for solving of the momentum equation, the rest of
        the cross section is considered to be only storage area in the
        continuity equation. By default, the whole cross section area
        is flowing.

        Avoid defining storage areas in the main channel as this can
        cause instability if only the non-flowing part of the channel
        is wetted.

        Additionally, the non-flowing areas are not taken into account
        for sediment deposition and erosion, by default None
        .. warning:: Mutually exclusive with
            :attr:`water_flow_slice_indexes`
        .. seealso:: :attr:`water_flow_slice_indexes`
    water_flow_slice_indexes : Tuple[int, int], optional
        Like :attr:`water_flow_range`, but uses edge indexes instead of
        coordinates, by default None
        .. warning:: Mutually exclusive with :attr:`water_flow_range`
        .. seealso:: :attr:`water_flow_range`
    main_channel_range : Tuple[float, float], optional
        Range defining the span of the main channel. Everything outside
        of the main channel is treated as floodplain. If not defined,
        the whole cross section is assumed to belong to the main
        channel, by default None
        .. warning:: Mutually exclusive with
            :attr:`main_channel_slice_indexes`
        .. seealso:: :attr:`main_channel_slice_indexes`
    main_channel_slice_indexes : Tuple[int, int], optional
        Like :attr:`main_channel_range`, but uses edge indexes instead
        of coordinates, by default None
        .. warning:: Mutually exclusive with :attr:`main_channel_range`
        .. seealso:: :attr:`main_channel_range`
    bottom_range : Tuple[float, float], optional
        Range defining the span of the bottom/bedload active zone, by
        default None
        .. warning:: Mutually exclusive with
            :attr:`bottom_slice_indexes`
        .. seealso:: :attr:`bottom_slice_indexes`
    bottom_slice_indexes : Tuple[float, float], optional
        Like :attr`bottom_range`, but uses edge indexes instead of
        coordinates, by default None
        .. warning:: Mutually exclusive with :attr:`bottom_range`
        .. seealso:: :attr:`bottom_range`
    active_range : Tuple[float, float], optional
        This range defines the active part of the cross section and
        must span from the left to the right dike. If not specified,
        the whole cross section is assumed to be active, and the
        elevation of the first and last point will be assumed to
        correspond to the dike elevation, by default None
        .. warning:: Mutually exclusive with
            :attr:`active_slice_indexes`
        .. seealso:: :attr:`active_slice_indexes`
    active_slice_indexes : Tuple[int, int], optional
        Like :attr:`active_range`, but uses edge indexes instead of
        coordinates, by default None
        .. warning:: Mutually exclusive with :attr:`active_range`
        .. seealso:: :attr:`active_range`
    bed_form_factor : float, optional
        Factor considering the influence of bed forms on bottom shear
        stress (ripple factor). This factor reduces the bottom shear
        stress due to bedforms with a constant value. If this factor is
        set to 1.0 (default), there is no influence and the standard
        shear stress is used, by default 1.0
        .. seealso:: :attr:`theta_critic`
    theta_critic : float, optional
        User-defined dimensionless critical bottom shear stress
        (Shields parameter) for this cross section. If specified, this
        local theta critic is used for this cross section. Permissible
        values range from 0.0 to 1.0. Set to a negative value to use
        the standard theta critical as defined in the
        ``BEDLOAD_PARAMETER`` block, by default -1.0
        .. seealso:: :attr:`bed_form_factor`
    friction_calibration_factor : float, optional
        Calibration factor between 0.0 and 10.0 to locally modify the
        friction value for this cross section, by default 1.0
    default_friction : float, optional
        Default Strickler friction value to use for this cross section,
        by default None
    friction_coefficients : List[float], optional
        Strickler values to apply to the segments in the groups defined
        in :attr:`friction_ranges` or :attr:`friction_slice_indexes`
        respectively, by default None
    friction_ranges : List[Tuple[float, float]], optional
        The ranges to which to apply the coefficients defined in
        :attr:`friction_coefficients`, by default None
        .. warning:: Mutually exclusive with
            :attr:`friction_slice_indexes`
        .. seealso:: :attr:`friction_slice_indexes`
    friction_slice_indexes : List[Tuple[int, int]], optional
        Like :attr:`friction_ranges`, but uses edge indexes instead of
        coordinates, by default None
        .. warning:: Mutually exclusive with :attr:`friction_ranges`
        .. seealso:: :attr:`friction_ranges`
    table_fixed_points : List[float], optional
        Fixed points which should be included in the table calculation.
        Note that these values are only used if you use lookup tables
        for hydraulic properties (see ``SECTION_COMPUTATION`` in the
        BASEMENT command file documentation for details), by default
        None
    table_fixed_lower_limit : float, optional
        Allows to fix the lowest water surface elevation used in the
        lookup table for ``A(z)`` to a different value than the lowest
        elevation in the cross section. This option is useful if you
        include a deep lake in your model containing a large water
        volume not participating in the dynamics of your river system,
        by default None
    interpolated : bool, optional
        This is an internal tag used to indicate that the cross section
        is the result of an interpolation of two user-defined cross
        sections, by default False
    soil_defs : List[BaseChainSoilDef], optional
        A list of soil definition blocks for the given cross section.
        See the :class:`BaseChainSoilDef` class for details, by default
        None

    """

    name: str
    distance_coord: float
    node_coords: List[Tuple[float, float]]
    sternberg: float = 0.0
    reference_height: float = 0.0
    left_point_global_coords: Optional[Tuple[float, float, float]] = None
    orientation_angle: float = 0.0
    interpolation_fixpoints: Optional[List[int]] = None
    water_flow_range: Optional[Tuple[float, float]] = None
    water_flow_slice_indexes: Optional[Tuple[int, int]] = None
    main_channel_range: Optional[Tuple[float, float]] = None
    main_channel_slice_indexes: Optional[Tuple[int, int]] = None
    bottom_range: Optional[Tuple[float, float]] = None
    bottom_slice_indexes: Optional[Tuple[int, int]] = None
    active_range: Optional[Tuple[float, float]] = None
    active_slice_indexes: Optional[Tuple[int, int]] = None
    bed_form_factor: float = 1.0
    theta_critic: float = -1.0
    friction_calibration_factor: float = 1.0
    default_friction: Optional[float] = None
    friction_coefficients: Optional[List[float]] = None
    friction_ranges: Optional[List[Tuple[float, float]]] = None
    friction_slice_indexes: Optional[List[Tuple[int, int]]] = None
    table_fixed_points: Optional[List[float]] = None
    table_fixed_lower_limit: Optional[float] = None
    interpolated: bool = False
    soil_defs: Optional[List[BaseChainSoilDef]] = None


class BaseChainReader:
    """Reader method for BMG files."""

    def __init__(self, filepath: str) -> None:
        self.filepath = filepath
        self._file: Optional[IO[str]] = None

    def __enter__(self) -> 'BaseChainReader':
        self._file = open(self.filepath)
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]],  # type: ignore
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> bool:
        self.close()
        return False

    def close(self) -> None:
        """Close the file.

        You do not need to call this method when using the parser as a
        context manager. It is only required if you manually called
        :meth:`BaseChainReader` earlier.
        """
        if self._file is not None:
            self._file.close()

    @staticmethod
    def _convert_values(data: Dict[str, Any]) -> BaseChainCrossSection:
        """Convert the string literals read from file to proper values.

        This is specific to the :class:`CrossSection` data class and
        will ignore any unexpected keys. The provided dictionary will
        get altered by this method.
        """
        # This table is used to look up the conversion type for a given key
        types: Dict[str, List[str]] = {
            'float': [
                'distance_coord', 'sternberg', 'reference_height',
                'orientation_angle', 'bed_form_factor', 'theta_critic',
                'friction_calibration_factor', 'default_friction',
                'table_fixed_lower_limit'],
            'str': [
                'name'],
            'tuple_2_float': [
                'water_flow_range',
                'main_channel_range',
                'bottom_range',
                'active_range'
            ],
            'tuple_2_int': [
                'water_flow_slice_indexes', 'main_channel_slice_indexes',
                'bottom_slice_indexes', 'active_slice_indexes'
            ],
            'tuple_3_float': [
                'left_point_global_coords'],
            'list_tuples_2_float': [
                'node_coords', 'friction_ranges'],
            'list_int': [
                'interpolation_fixpoints'],
            'list_float': [
                'friction_coefficients', 'table_fixed_points'],
            'list_tuples_2_int': [
                'friction_slice_indexes'],
            'bool': [
                'interpolated']
        }

        for key, value in data.items():
            if key in types['float']:
                data[key] = float(value)
            elif key in types['str']:
                data[key] = str(value)
            elif key in types['tuple_2_float']:
                value = f'{value[:-1]},)'
                tuple_ = ast.literal_eval(value)
                assert isinstance(tuple_, tuple), f'Unexpected literal {value}'
                assert len(tuple_) == 2, f'Unexpected tuple size for {key}'
                data[key] = tuple(float(e) for e in tuple_)
            elif key in types['tuple_3_float']:
                value = f'{value[:-1]},)'
                tuple_ = ast.literal_eval(value)
                assert isinstance(tuple_, tuple), f'Unexpected literal {value}'
                assert len(tuple_) == 3, f'Unexpected tuple size for {key}'
                data[key] = tuple(float(e) for e in tuple_)
            elif key in types['tuple_2_int']:
                value = f'{value[:-1]},)'
                tuple_ = ast.literal_eval(value)
                assert isinstance(tuple_, tuple), f'Unexpected literal {value}'
                assert len(tuple_) == 2, f'Unexpected tuple size for {key}'
                data[key] = tuple(int(e) for e in tuple_)
            elif key in types['list_tuples_2_float']:
                value = f'{value[:-1]},)'
                tuple_ = ast.literal_eval(value)
                assert isinstance(tuple_, tuple), f'Unexpected literal {value}'
                assert len(tuple_[0]) == 2, f'Unexpected tuple size for {key}'
                data[key] = [tuple(float(e) for e in t) for t in tuple_]
            elif key in types['list_tuples_2_int']:
                value = f'{value[:-1]},)'
                tuple_ = ast.literal_eval(value)
                assert isinstance(
                    tuple_[0], tuple), f'Unexpected literal {value}'
                assert len(tuple_[0]) == 2, f'Unexpected tuple size for {key}'
                data[key] = [tuple(int(e) for e in t) for t in tuple_]
            elif key in types['list_int']:
                value = f'{value[:-1]},)'
                tuple_ = ast.literal_eval(value)
                assert isinstance(tuple_, tuple), f'Unexpected literal {value}'
                data[key] = [int(e) for e in tuple_]
            elif key in types['list_float']:
                value = f'{value[:-1]},)'
                tuple_ = ast.literal_eval(value)
                assert isinstance(tuple_, tuple), f'Unexpected literal {value}'
                data[key] = [float(e) for e in tuple_]
            elif key in types['bool']:
                data[key] = bool(value)

        return BaseChainCrossSection(**data)

    def cross_sections(self) -> Iterator[BaseChainCrossSection]:
        """Iterate over the file returning any cross sections found."""
        if self._file is None:
            raise RuntimeError('Cannot iterate over closed file')

        # This variable holds the last few lines of the file
        cache: str = ''
        for line in self._file:
            # Handle comments
            line, *_ = line.split('//', maxsplit=1)
            cache += ''.join(line.strip(' \t'))

        offset = 0
        while 'CROSS_SECTION' in cache:
            start, end = match_tokens(cache[offset:], '{', '}')
            if start == -1 and end == -1:
                break
            yield self._process_block(cache[start:end])
            cache = cache[end:]

    @staticmethod
    def _parse_soildef(string: str) -> BaseChainSoilDef:
        """Process the given soil definition substring."""
        assert string.startswith('{')
        assert string.endswith('}')
        string = string[1:-1].strip()
        data: Dict[str, Any] = {}
        for line in string.splitlines():
            key, _, *values = line.split()
            value = ''.join(values)
            if key == 'index':
                data[key] = int(value)
            elif key == 'range':
                tuple_ = ast.literal_eval(value)
                data[key] = tuple(float(i) for i in tuple_)
            elif key == 'slice_indexes':
                tuple_ = ast.literal_eval(value)
                data[key] = tuple(int(i) for i in tuple_)

        return BaseChainSoilDef(**data)

    def _process_block(self, string: str) -> BaseChainCrossSection:
        """Process a given substring to extract a cross section.

        The provided string will be modified in the process.
        """
        assert string.startswith('{')
        assert string.endswith('}')
        string = string[1:-1].strip()

        # This will store the key-value pairs read from the file
        data: Dict[str, Any] = {}

        # Extract any SOIL_DEF keys
        while True:
            if not '{' in string:
                break
            sd_start, sd_end = match_tokens(string, '{', '}')
            # Extract the tag name
            pre = string[:sd_start]
            pre, tag = pre.strip().rsplit(maxsplit=1)
            # Extract soil definitions
            if tag == 'SOIL_DEF':
                soil_def = self._parse_soildef(string[sd_start:sd_end])
                try:
                    data['soil_defs'].append(soil_def)
                except KeyError:
                    data['soil_defs'] = [soil_def]
            else:
                warnings.warn(f'Ignoring unknown tag {tag}')
            string = pre + string[sd_end:]

        while True:
            # NOTE: This uses equal signs to separate two arguments. I found
            # this to be easier/faster than parsing the parentheses due to the
            # overhead involved.

            # Get the first two equal signs in the string
            first = string.index('=')
            try:
                second: Optional[int] = string.index('=', first+1)
            except ValueError:
                second = None

            # The substring before the first is the name of the key
            key = string[:first].strip()

            # The "body" is the chunk between the two equal signs, including
            # the next key (will be trimmed later).
            if second is None:
                inner = body = string[first+1:]
                next_key = None
            else:
                body = string[first+1:second]
                inner, next_key = body.rsplit('\n', maxsplit=1)

            # The value of a key may be serialised as newlines no longer act as
            # separators
            data[key] = ''.join(inner.split())

            if next_key is None:
                break

            string = string[string.index(next_key):]

        return self._convert_values(data)

    def open(self, **kwargs: Any) -> IO[str]:
        """Manually open the wrapped file.

        It is recommended to use the parser as a context manager
        instead. If you do use this method, make sure to call
        :meth:`BaseChainReader.close()` afterwards.
        """
        self._file = open(self.filepath, **kwargs)
        assert self._file is not None
        return self._file


class BaseChainWriter:
    """Writer class for BMG files."""

    def __init__(self, filepath: str,
                 sections: Optional[Iterable[BaseChainCrossSection]] = None
                 ) -> None:
        self._sections: List[BaseChainCrossSection] = list(sections or ())
        self.filepath = filepath
        self._file: Optional[IO[str]] = None

        if sections:
            self._write_header()
            for section in sections:
                self._write_section(section)

    def __enter__(self) -> 'BaseChainWriter':
        self._file = self.open('w')
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]],  # type: ignore
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> bool:
        self.close()
        return False

    def close(self) -> None:
        """Close the file.

        You do not need to call this method when using the parser as a
        context manager. It is only required if you manually called
        :meth:`BaseChainReader` earlier.
        """
        if self._file is not None:
            self._file.close()

    def open(self, *args: Any, **kwargs: Any) -> IO[str]:
        """Manually open the wrapped file.

        It is recommended to use the writer as a context manager
        instead. If you do use this method, make sure to call
        :meth:`BaseChainWriter.close()` afterwards.
        """
        self._file = open(self.filepath, *args, **kwargs)
        assert self._file is not None
        return self._file

    def write_channel(self, channel: 'ChannelGeometry') -> None:
        """Write a channel geometry to disk."""
        self._write_header()
        sections = channel.to_bmg()
        for index, section in enumerate(sections):
            self._write_section(section)
            if not index+1 == len(sections):
                self._file.write('\n')  # type: ignore

    def _write_header(self) -> None:
        # pylint: disable=import-outside-toplevel
        from . import __version__ as module_version
        now = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        assert self._file is not None
        self._file.write(f'// BASEMENT 1D Geometry\n// Generated {now} using '
                         f'BASEchange v{module_version}\n')

    def _write_section(self, section: BaseChainCrossSection) -> None:
        assert self._file is not None
        # Required tokens
        text = [
            'CROSS_SECTION {',
            f'\tname = {section.name}',
            f'\tdistance_coord = {section.distance_coord}',
            '\tnode_coords = '
            + '(' + ','.join(
                (_compact_list(n) for n in section.node_coords)) + ')']
        # Optional tags
        if section.sternberg != 0.0:
            text.append(f'\tsternberg = {section.sternberg}')
        if section.reference_height != 0.0:
            text.append(f'\treference_height = {section.reference_height}')
        if section.left_point_global_coords is not None:
            text.append('\tleft_point_global_coords = '
                        + _compact_list(section.left_point_global_coords))
        if section.orientation_angle != 0.0:
            text.append(f'\torientation_angle = {section.orientation_angle}')
        if section.interpolation_fixpoints is not None:
            text.append('\tinterpolation_fixpoints = '
                        f'{_compact_list(section.interpolation_fixpoints)}')
        if section.water_flow_range is not None:
            text.append('\twater_flow_range = '
                        f'{_compact_list(section.water_flow_range)}')
        if section.water_flow_slice_indexes is not None:
            text.append('\twater_flow_slice_indexes = '
                        f'{_compact_list(section.water_flow_slice_indexes)}')
        if section.main_channel_range is not None:
            text.append(f'\tmain_channel_range = '
                        f'{_compact_list(section.main_channel_range)}')
        if section.main_channel_slice_indexes is not None:
            text.append(f'\tmain_channel_slice_indexes = '
                        f'{_compact_list(section.main_channel_slice_indexes)}')
        if section.bottom_range is not None:
            text.append(f'\tbottom_range = '
                        f'{_compact_list(section.bottom_range)}')
        if section.bottom_slice_indexes is not None:
            text.append(f'\tbottom_slice_indexes = '
                        f'{_compact_list(section.bottom_slice_indexes)}')
        if section.active_range is not None:
            text.append(f'\tactive_range = '
                        f'{_compact_list(section.active_range)}')
        if section.active_slice_indexes is not None:
            text.append(f'\tactive_slice_indexes = '
                        f'{_compact_list(section.active_slice_indexes)}')
        if section.bed_form_factor != 1.0:
            text.append(f'\tbed_form_factor = {section.bed_form_factor}')
        if section.theta_critic != -1.0:
            text.append(f'\ttheta_critic = {section.theta_critic}')
        if section.friction_calibration_factor != 1.0:
            text.append('\tfriction_calibration_factor = '
                        f'{section.friction_calibration_factor}')
        if section.default_friction is not None:
            text.append(f'\tdefault_friction = {section.default_friction}')
        if section.friction_coefficients is not None:
            text.append(f'\tfriction_coefficients = '
                        f'{_compact_list(section.friction_coefficients)}')
        if section.friction_ranges is not None:
            line = '\tfriction_ranges = ('
            line += ','.join(
                (_compact_list(r) for r in section.friction_ranges))
            line += ')'
            text.append(line)
        if section.friction_slice_indexes is not None:
            line = '\tfriction_slice_indexes = ('
            line += ','.join(
                (_compact_list(r) for r in section.friction_slice_indexes))
            line += ')'
            text.append(line)
        if section.table_fixed_points is not None:
            text.append('\ttable_fixed_points = '
                        f'{_compact_list(section.table_fixed_points)}')
        if section.table_fixed_lower_limit is not None:
            text.append('\ttable_fixed_lower_limit = '
                        f'{section.table_fixed_lower_limit}')
        if section.soil_defs is not None:
            for def_ in section.soil_defs:
                line = ('\tSOIL_DEF {\n'
                        f'\t\tindex = {def_.index}\n')
                if def_.slice_indexes is not None:
                    line += ('\t\tslice_indexes = '
                             f'{_compact_list(def_.slice_indexes)}')
                elif def_.range is not None:
                    line += f'\t\trange = {_compact_list(def_.range)}'
                line += '\n\t}'
                text.append(line)
        text.append('}')
        out_file: IO[str] = self._file
        out_file.write('\n'.join(text))


def match_tokens(text: str, start: str, end: Optional[str] = None
                 ) -> Tuple[int, int]:
    """Extract the substring enclosed between the two tokens.

    This scans the given string for the start token and extracts the
    start and end indices of the enclosed substring. The tokens are not
    included in this substring.

    This will return ``(-1, -1)`` if neither token is found.

    If multiple nested pairs of start and end tokens are found, the
    inner tokens must be matched before the outer tokens. For example,
    in the string ``animal ( feline ( cat ) )``, the substring matched
    for the parentheses would be `` feline ( cat ) ``.

    This method is intended for non-serialisable, multiline strings
    potentially containing nested sets of matching tokens, performance
    for short strings is likely not great compared to other solutions
    like :meth:`str.find()`` or regular expressions.

    Parameters
    ----------
    text : str
        The input text to process.
    start : str
        The opening token to scan for.
    end : str or None, optional
        The closing token. If None, the starting token will be used,
        by default None

    Returns
    -------
    Tuple[int, int]
        The start and end index of the matching substring in slice
        notation, or ``(-1, -1)`` if neither was found.

    Raises
    ------
    TokenError
        Raised if a mismatch between an opening and closing token is
        encountered.

    """
    if end is None:
        end = start
    match_start = match_end = -1
    # This variable keeps track of any nested sets of tokens. 0 means no match
    # yet, higher numbers indicate the current depth.
    nested = 0

    chunk = text  # The active substring that is searched for tokens
    chunk_offset = 0  # Conversion between the global index and the chunk
    while chunk:
        next_ = ''

        # Find the next token in the active substring
        if start in chunk:
            next_ = start
            index = chunk.index(start)
            # If there is an end token before the start token, use that instead
            if end in chunk and chunk.index(end) < index:
                next_ = end
                index = chunk.index(end)
        elif end in chunk:
            next_ = end
            index = chunk.index(end)

        if next_ == start:
            # The first starting token found automatically becomes the start of
            # the match
            if not nested:
                match_start = chunk_offset + index
            chunk_offset += index + len(start)
            nested += 1

        elif next_ == end:
            # Closing tokens with no matching opening token are not permitted
            if not nested:
                raise TokenError(
                    f'Unexpected closing token {end} encountered; no matching '
                    'opening token found')
            chunk_offset += index + len(end)
            nested -= 1
            if nested < 1:
                match_end = chunk_offset
                break
        # Advance the current chunk
        chunk = text[chunk_offset:]
    # Check for unmatched end tokens
    if match_start != -1 and match_end == -1:
        raise TokenError(f'Unmatched opening token {start}')
    return match_start, match_end


def _compact_list(data: Sequence[Any], inline: bool = True,
                  newline_indent: str = '\t') -> str:
    """Write a tuple without spaces."""
    text = '('
    if inline:
        text += ','.join((str(d) for d in data))
    else:
        text += ''.join(f'{newline_indent}{d},\n' for d in data)
        text += newline_indent
    text += ')'
    return text



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Defines the hole and region point markers for quality meshing."""

from typing import Optional, Tuple


class HoleMarker:
    """Marks a segment-bounded region as a hole."""

    __slots__ = ('_pos_x', '_pos_y')

    def __init__(self, pos_x: float, pos_y: float) -> None:
        """Initialize a new hole marker.

        Parameters
        ----------
        pos_x : float
            X position of the marker
        pos_y : float
            Y position of the marker
        """
        self._pos_x = pos_x
        self._pos_y = pos_y

    def as_tuple(self) -> Tuple[float, float]:
        """Return the coordinates of the marker as a tuple."""
        return self._pos_x, self._pos_y


class RegionMarker:
    """Defines properties of a segment-bounded region."""

    __slots__ = ('_attribute', '_max_area', '_pos_x', '_pos_y')

    def __init__(self, pos_x: float, pos_y: float, *, max_area: float = -1.0,
                 attribute: Optional[int] = None) -> None:
        """Initialise a new region marker.

        Parameters
        ----------
        pos_x : float
            X position of the marker
        pos_y : float
            Y position of the marker
        max_area : float, optional
            A maximum area constraint to apply to elements in this
            region, negative values signify that no area constraints
            should be applied, by default -1.0
        attribute : int, optional
            A regional attribute to associate with elements in this
            region, by default None
        """
        self._attribute = attribute
        self._max_area = max_area if max_area > 0.0 else -1.0
        self._pos_x = pos_x
        self._pos_y = pos_y

    @property
    def attribute(self) -> Optional[int]:
        """Return the attribute to apply to elements in this region."""
        return self._attribute

    @property
    def max_area(self) -> float:
        """Return the maximum area constraint for this region.

        A value of -1.0 signifies that no constraint will be applied.
        """
        return self._max_area

    def as_tuple(self) -> Tuple[float, float]:
        """Return the coordinates of the marker as a tuple."""
        return self._pos_x, self._pos_y



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Core components of the mesh generator."""

from abc import ABCMeta, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple
from ..abc import ElevationSource
from ..core import Mesh
from ..core.geometry import Node, Segment
from ..interpolation import interpolate_mesh
from ..triangle import HoleMarker, RegionMarker, PSLGNode, PSLGSegment
from ..types import Point2D
from ..meshing import quality_mesh
from ..stringdefs import resolve_string_defs


class AbstractFactory(ElevationSource, metaclass=ABCMeta):
    """Base class for mesh factories.

    This provides two main endpoints for generating meshes. The first
    is to subclass this method and override the elevation_at name
    with a custom method.

    The second is to assign any callable as the elevation function
    using the set_elevation_func method. The callable provided must
    have the same signature as the elevation_at method itself
    (Tuple[float, float] -> float).
    """

    def __init__(self) -> None:
        """Initialise the factory.

        This does not perform any setup, but calling it is recommended
        regardless.
        """
        self.nodes: List[Node] = []
        self.segments: List[Segment] = []

        self.holes: Optional[List[HoleMarker]] = None
        self.regions: Optional[List[RegionMarker]] = None
        self.string_defs: Optional[Dict[str, List[Node]]] = None

    def build(self, max_area: float = -1.0, min_angle: float = 28.0,
              **kwargs: Any) -> Tuple[Mesh, Dict[str, List[int]]]:
        """Generate and return a mesh using the given inputs.

        This first calls 'Triangle' to generate quality mesh before
        repeatedly calling the elevation function and interpolating
        the mesh nodes.

        Parameters
        ----------
        max_area : float, optional
            The maximum are of any generated elements, by default -1.0
        min_angle : float, optional
            The minimum angle of any generated elements, by
            default 28.0

        Returns
        -------
        Tuple[Mesh, Dict[str, List[int]]]
            The generated mesh, as well as a list of string definitions
            mapped to their respective node IDs.
        """
        # Generate input geometry
        triangle_nodes = [PSLGNode(*n.as_tuple_2d()) for n in self.nodes]
        triangle_segments = [
            PSLGSegment.from_points(*l.as_line(), nodes=triangle_nodes)
            for l in self.segments]
        # Run triangle
        mesh = quality_mesh(triangle_nodes, triangle_segments,
                            self.holes or [], self.regions,
                            min_angle=min_angle, max_area=max_area, **kwargs)

        # Process string definitions
        if self.string_defs is not None:
            # NOTE: The string defs are provided by the user as named lists of
            # nodes, but the parser expects point tuples.
            # This is converted here.
            sd_points = {name: tuple(n.as_tuple_2d() for n in line_string)
                         for name, line_string in self.string_defs.items()}
            string_defs = resolve_string_defs(sd_points, mesh, 1e-6)
        else:
            string_defs = {}

        # Interpolate mesh
        # NOTE: This works because the MeshFactory class itself supports the
        # ElevationSource ABC requried for interpolation
        interpolate_mesh(mesh, self, default=-999.0)
        return mesh, string_defs

    @abstractmethod
    def elevation_at(self, point: Point2D) -> float:
        """Return the elevation of the given point.

        This method must be overridden by the user.

        Parameters
        ----------
        point : Tuple[float, float]
            The point to interpolate

        Returns
        -------
        float
            The elevation at the given point

        Raises
        ------
        NotImplementedError
            Always raised as this is an abstract method
        """
        raise NotImplementedError


class BasicFactory(AbstractFactory):
    """Basic mesh factory.

    This is a helper factory that allows the user to inject a custom
    elevation function at runtime using the set_elevation_func
    method.

    Overriding the original elevation_at method is still supported.
    """

    def __init__(self) -> None:
        """Initialise the mesh factory.

        This initialiser is provided for neatness and analysis reasons,
        but actually calling it is not required for execution.
        """
        super().__init__()
        self.elevation_func: Callable[[Point2D], float]

    def elevation_at(self, point: Point2D) -> float:
        """Return the elevation at a given point.

        This calls the elevation function specified using the
        set_elevation_func decorator.

        Parameters
        ----------
        point : Tuple[float, float]
            The point to interpolate

        Returns
        -------
        float
            The interpolated elevation of the point

        Raises
        ------
        RuntimeError
            Raised if no elevation function has been specified by the
            user
        """
        try:
            return self.elevation_func(point)
        except AttributeError as err:
            raise RuntimeError('No elevation function defined') from err

    def set_elevation_func(self, func: Callable[[Point2D], float]) -> None:
        """Set the elevation function to use for this factory.

        This is mostly a syntactic shorthand; the following two
        statements are effectively identical:

            factory = MeshFactory()

            # Regular function assignment assignment
            def foo(point):
                return abs(point[0] + point[1])
            factory.set_elevation_func(foo)

            # Decorator shorthand
            factory = MeshFactory()
            @factory.set_elevation_func
            def foo(point):
                return abs(point[0] + point[1])

        Parameters
        ----------
        func : Callable[[Tuple[float, float]], float]
            A user-provided elevation function used to generate the
            elevation at a given point
        """
        self.elevation_func = func



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Algorithms related to MATID visualisation and modification."""

from typing import Any, Dict, List, Optional, Tuple
from PyQt5.QtCore import QVariant
from qgis import processing  # type: ignore
from qgis.core import (QgsFeature, QgsFeatureSink, QgsField,  # type: ignore
                       QgsFields, QgsGeometry, QgsMeshLayer, QgsPointXY,
                       QgsProcessingAlgorithm,
                       QgsProcessingParameterDefinition,
                       QgsProcessingParameterFeatureSink,
                       QgsProcessingParameterField,
                       QgsProcessingParameterFileDestination,
                       QgsProcessingParameterMeshLayer,
                       QgsProcessingParameterVectorLayer, QgsProcessing,
                       QgsProcessingContext, QgsProcessingFeedback,
                       QgsTriangle, QgsVectorLayer, QgsWkbTypes)
from ...core.mesh import Mesh, MeshElement
from ...types import Polygon2D
from .utils import parse_docstring


class AssignMaterialID(QgsProcessingAlgorithm):
    """Assign element MATIDs based on containment in a polygon."""

    # Parameter literals

    INPUT_MESH = 'INPUT_MESH'
    INPUT_POLYGON = 'INPUT_POLYGON'
    MATID_FIELD = 'MATID_FIELD'
    OUTPUT = 'OUTPUT'

    # Constructor

    @classmethod
    def createInstance(cls) -> 'AssignMaterialID':
        """Return a new copy of the algorithm."""
        return cls()

    # QGIS properties

    @staticmethod
    def name() -> str:
        """Return the unique algorithm name."""
        return 'assignmaterialid'

    @staticmethod
    def displayName() -> str:
        """Return the user-facing algorithm name."""
        return 'Assign MATID'

    @classmethod
    def shortHelpString(cls) -> str:
        """Return a short help string for the algorithm."""
        return parse_docstring(cls.__doc__)

    def initAlgorithm(self, _: Optional[Dict[str, Any]] = None) -> None:
        """Define the algorithm's inputs and outputs."""
        # Input mesh
        self.addParameter(QgsProcessingParameterMeshLayer(
            name=self.INPUT_MESH,
            description='Mesh layer'))
        # Input polygon
        self.addParameter(QgsProcessingParameterVectorLayer(
            name=self.INPUT_POLYGON,
            description='Input polygons',
            types=[QgsProcessing.TypeVectorPolygon]))
        # Polygon MATID field
        self.addParameter(QgsProcessingParameterField(
            name=self.MATID_FIELD,
            description='MATID field to apply to the mesh',
            parentLayerParameterName=self.INPUT_POLYGON,
            type=QgsProcessingParameterField.Numeric))
        # Output
        self.addParameter(QgsProcessingParameterFileDestination(
            name=self.OUTPUT,
            description='Output mesh layer',
            fileFilter='SMS 2D mesh (*2dm)'))

    def processAlgorithm(self, parameters: Dict[str, Any],
                         context: QgsProcessingContext,
                         feedback: QgsProcessingFeedback) -> Dict[str, Any]:
        """Execute the algorithm using the given parameters.

        Parameters
        ----------
        parameters : Dict[str, Any]
            The parameters defined for this algorithm
        context : QgsProcessingContext
            The context in which the algorithm is run
        feedback : QgsProcessingFeedback
            The feedback object to use for status reporting

        Returns
        -------
        Dict[str, Any]
            The algorithms output parameters
        """
        input_mesh_layer: QgsMeshLayer = self.parameterAsMeshLayer(
            parameters, self.INPUT_MESH, context)
        mesh = Mesh.open(input_mesh_layer.source())
        input_polygon_layer: QgsVectorLayer = self.parameterAsVectorLayer(
            parameters, self.INPUT_POLYGON, context)
        input_field: QgsField = self.parameterAsFields(
            parameters, self.MATID_FIELD, context)[0]
        output_file = self.parameterAsFileOutput(
            parameters, self.OUTPUT, context)

        # Extract polygons
        polygons: List[Tuple[int, Polygon2D]] = []
        feature: QgsFeature
        for feature in input_polygon_layer.getFeatures():
            attr_index = feature.fields().indexFromName(input_field)
            matid = int(feature.attributes()[attr_index])
            polygon = []
            for vertex in feature.geometry().vertices():
                polygon.append((vertex.x(), vertex.y()))
            polygons.append((matid, polygon))  # type: ignore

        # Modify and save the mesh
        elements: List[MeshElement]
        for matid, polygon in polygons:  # type: ignore
            elements = mesh.elements_by_polygon(polygon)  # type: ignore
            for element in elements:
                element.materials = (matid,)
        mesh.save(output_file)

        feedback.setProgress(100)
        return {self.OUTPUT: output_file}


class ExtractMaterialID(QgsProcessingAlgorithm):
    """Create MultiPolygon geometries from the mesh Material IDs.

    This will group the mesh elements by their assigned MATID and write
    a MultiPolgon layer with the corresponding MATID as its attribute.
    """

    # Parameter literals

    INPUT = 'INPUT'
    _TEMP = 'TEMP_OUTPUT'
    OUTPUT = 'OUTPUT'

    # Constructor

    @classmethod
    def createInstance(cls) -> 'ExtractMaterialID':
        """Return a new copy of the algorithm."""
        return cls()

    # QGIS properties

    @staticmethod
    def name() -> str:
        """Return the unique algorithm name."""
        return 'extractmaterialid'

    @staticmethod
    def displayName() -> str:
        """Return the user-facing algorithm name."""
        return 'Extract MATID'

    @classmethod
    def shortHelpString(cls) -> str:
        """Return a short help string for the algorithm."""
        return parse_docstring(cls.__doc__)

    def initAlgorithm(self, _: Optional[Dict[str, Any]] = None) -> None:
        """Define the algorithm's inputs and outputs."""
        # Input layer
        self.addParameter(QgsProcessingParameterMeshLayer(
            name=self.INPUT,
            description='Mesh layer'))
        # Temporary sink
        temp_sink = QgsProcessingParameterFeatureSink(
            name='TEMP_OUTPUT',
            description='Internal sink for intermediate algorithms',
            type=QgsProcessing.TypeVectorPolygon,
            defaultValue='memory:',
            optional=True)
        hidden_flag = QgsProcessingParameterDefinition.FlagHidden
        temp_sink.setFlags(temp_sink.flags() | hidden_flag)
        self.addParameter(temp_sink)
        # Output
        self.addParameter(QgsProcessingParameterFeatureSink(
            name=self.OUTPUT,
            description='Output polygon layer',
            type=QgsProcessing.TypeVectorPolygon))

    def processAlgorithm(self, parameters: Dict[str, Any],
                         context: QgsProcessingContext,
                         feedback: QgsProcessingFeedback) -> Dict[str, Any]:
        """Execute the algorithm using the given parameters.

        Parameters
        ----------
        parameters : Dict[str, Any]
            The parameters defined for this algorithm
        context : QgsProcessingContext
            The context in which the algorithm is run
        feedback : QgsProcessingFeedback
            The feedback object to use for status reporting

        Returns
        -------
        Dict[str, Any]
            The algorithms output parameters
        """
        MATID_FIELD = 'MATID'
        COUNT_FIELD = 'Element count'

        mesh_layer: QgsMeshLayer = self.parameterAsMeshLayer(
            parameters, self.INPUT, context)
        mesh = Mesh.open(mesh_layer.source())

        fields = QgsFields()
        fields.append(QgsField(MATID_FIELD, type=QVariant.Int))
        fields.append(QgsField(COUNT_FIELD, type=QVariant.Int))
        temp_sink, temp_dest_id = self.parameterAsSink(
            parameters, self._TEMP, context, fields=fields,
            geometryType=QgsWkbTypes.MultiPolygon, crs=mesh_layer.crs())

        # Group mesh elements by MATID
        grouped: Dict[int, List[MeshElement]] = {}
        for element in mesh.elements:
            matid = int(element.materials[0])
            try:
                grouped[matid].append(element)
            except KeyError:
                grouped[matid] = [element]

        # Collect geometries for all elements
        for matid, elements in grouped.items():
            element_count = len(elements)
            for element in elements:
                geom = QgsGeometry(QgsTriangle(
                    *[QgsPointXY(*p[:2]) for p in element.points]))
                feat = QgsFeature()
                feat.setGeometry(geom)
                feat.setAttributes([matid, element_count])
                temp_sink.addFeature(feat, QgsFeatureSink.FastInsert)

        feedback.setProgress(50)

        # Dissolve layer
        dissolve_results = processing.run(
            'native:dissolve',
            {'INPUT': temp_dest_id,
             'FIELD': [MATID_FIELD],
             'OUTPUT': parameters[self.OUTPUT]},
            is_child_algorithm=True,
            context=context,
            feedback=feedback)

        return {self.OUTPUT: dissolve_results['OUTPUT']}
